/* --------------------------------------------------------------- *\

    file  : fbcmd.cpp
    author: m. gumz
    mail  : gumz at cs.uni-magdeburg.de

    start : Fri Jun 25 00:03:33 2004

\* --------------------------------------------------------------- */
/* --------------------------------------------------------------- *\

   Copyright (c) 2004 mathias gumz

  Permission is hereby granted, free of charge, to any person
  obtaining a copy of this software and associated documentation
  files (the "Software"), to deal in the Software without
  restriction, including without limitation the rights to use,
  copy, modify, merge, publish, distribute, sublicense, and/or
  sell copies of the Software, and to permit persons to whom
  the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be
  included in all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
  OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
  OTHER DEALINGS IN THE SOFTWARE.

\* --------------------------------------------------------------- */

/* --------------------------------------------------------------- *\
  includes
\* --------------------------------------------------------------- */
#include <iostream>
#include <string>
#include <vector>
#include <algorithm>
#include <sstream>

#include <unistd.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <ctype.h>

#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>
#include <X11/Xmu/WinUtil.h>

/*------------------------------------------------------------------*\
\*------------------------------------------------------------------*/

#define VERSION "0.5.1"

/*------------------------------------------------------------------*\
\*------------------------------------------------------------------*/

typedef std::vector< std::string > ArgList;
typedef std::vector< std::string > PathList;

typedef struct FbInfo{

  Display* display;
  int      screen_number;
  Window   root_win;

  FbInfo() : display(0), screen_number(0), root_win(0) { };
} FbInfo;

typedef struct FbCmdEntry{

  const char* cmdname;
  int (*command)(const FbInfo& fbinfo, const ArgList& arglist);

} FbCmdEntry;

typedef struct LogutEntry {
  const char* appname;
  const char* options;
} LogutEntry;

/*------------------------------------------------------------------*\
\*------------------------------------------------------------------*/

int cmdHelp(const FbInfo&, const ArgList&);
int cmdPid(const FbInfo&, const ArgList&);
int cmdVersion(const FbInfo&, const ArgList&);
int cmdLogout(const FbInfo&, const ArgList&);
int cmdPrint(const FbInfo&, const ArgList&);
int cmdWorkspace(const FbInfo&, const ArgList&);
int cmdNextWorkspace(const FbInfo&, const ArgList&);
int cmdPrevWorkspace(const FbInfo&, const ArgList&);

/*------------------------------------------------------------------*\
\*------------------------------------------------------------------*/

FbCmdEntry fbcommands[] = {
  { "help", cmdHelp },
  { "version", cmdVersion },
  { "logout", cmdLogout },
  { "print", cmdPrint},
  { "workspace", cmdWorkspace },
  { "ws", cmdWorkspace },
  { "nextworkspace", cmdNextWorkspace },
  { "nws", cmdNextWorkspace },
  { "prevworkspace", cmdPrevWorkspace },
  { "pws", cmdPrevWorkspace },
  { 0, 0 }
};

LogutEntry logouts[] = {

  { "xmessage", " -default -nearmouse -buttons no:1,yes:0 " },
  { "gxmessage", " -default -nearmouse -buttons no:1,yes:0 " },
  { "zenity", " --question --text=" },
  { "Xdialog", " --allow-close --under-mouse --no-ok --yesno 'do you really want to logout?' 10 50" },
  { "gdialog", " --yesno " },
  { "kdialog", " --yesno " },
  { 0, 0 }
};

/* --------------------------------------------------------------- *\
\* --------------------------------------------------------------- */

void displayUsage() {

  std::cout <<

"\n"
"ABOUT\n"
"\n"
"  fbcmd provides some functions to the 'core' of fluxbox\n"
"  from the commandline.\n"
"\n"
"USAGE\n"
"\n"
"  fbcmd action [arg1] [[arg2]] ...\n"
"\n"
"ACTIONS\n"
"\n"
"  help                      - display this help\n"
"  version                   - display version\n"
"  print                     - print info, try  print help\n"
"  nextworkspace [int]       - change to nextworkspace\n"
"  prevworkspace [int]       - change to prevworkspace\n"
"  workspace int             - change to workspace x\n"
"  logout [dialog [message]] - provide a logout-dialog\n"
"\n"
"AUTHOR\n"
"\n"
"  m. gumz (akira)\n"
"\n";

}

/*------------------------------------------------------------------*\
\*------------------------------------------------------------------*/

std::string toLower(const std::string& in) {
  std::string out = in;
  std::transform(out.begin(), out.end(), out.begin(), tolower);
  return out;
}


template <typename Container>
static void
stringToK (Container &container, std::string const &in,
           const char * const delimiters = " \t\n") {

    const std::string::size_type len = in.length();
    std::string::size_type i = 0;

    while ( i < len ) {
        // eat leading whitespace
        i = in.find_first_not_of(delimiters, i);
        if (i == std::string::npos)
            return;   // nothing left but white space

        // find the end of the token
        std::string::size_type j = in.find_first_of(delimiters, i);

        // push token
        if (j == std::string::npos) {
            container.push_back(in.substr(i));
            return;
        } else
            container.push_back(in.substr(i, j-i));

        // set up for next loop
        i = j + 1;
    }
}

void sendMessageToRootWindow(const FbInfo& fbinfo, Atom atom,
    unsigned long data1, unsigned long data2 = 0, unsigned long data3 = 0,
    unsigned long data4 = 0) {

  XEvent clientevent;
  clientevent.xclient.display = fbinfo.display;
  clientevent.xclient.type = ClientMessage;
  clientevent.xclient.window = fbinfo.root_win;
  clientevent.xclient.message_type = atom;
  clientevent.xclient.format = 32;
  clientevent.xclient.data.l[0] = data1;
  clientevent.xclient.data.l[1] = data2;
  clientevent.xclient.data.l[2] = data3;
  clientevent.xclient.data.l[3] = data4;

  XSendEvent(fbinfo.display,
             fbinfo.root_win,
             true,
             SubstructureRedirectMask,
             &clientevent);

}

Window selectWindow(const FbInfo& fbinfo) {
  int status;
  Cursor cursor;
  XEvent event;
  Window target_win = None;
  int buttons = 0;

  cursor = XCreateFontCursor(fbinfo.display, XC_crosshair);

  status = XGrabPointer(fbinfo.display, fbinfo.root_win, False,
                        ButtonPressMask|ButtonReleaseMask, GrabModeSync,
                        GrabModeAsync, fbinfo.root_win, cursor, CurrentTime);
  if (status != GrabSuccess) {
    std::cerr << "can't grab the mouse.";
    return None;
  }

  while ((target_win == None) || (buttons != 0)) {
    XAllowEvents(fbinfo.display, SyncPointer, CurrentTime);
    XWindowEvent(fbinfo.display, fbinfo.root_win, ButtonPressMask|ButtonReleaseMask, &event);
    switch (event.type) {
    case ButtonPress:
      if (target_win == None) {
        target_win = event.xbutton.subwindow;
        if (target_win == None)
          target_win = fbinfo.root_win;
      }
      buttons++;
      break;
    case ButtonRelease:
      if (buttons > 0)
        buttons--;
      break;
    }
  }

  XUngrabPointer(fbinfo.display, CurrentTime);

  if (target_win != None) {
    Window geometry_win;
    int dummyi;
    unsigned int dummy;

    if (XGetGeometry (fbinfo.display, target_win, &geometry_win, &dummyi, &dummyi,
          &dummy, &dummy, &dummy, &dummy) && target_win != geometry_win)
      target_win = XmuClientWindow(fbinfo.display, target_win);
  }

  return target_win;
}

std::string printAppsEntry(const FbInfo& fbinfo, const ArgList& list) {

  std::string result;
  Window win = None;
  win = selectWindow(fbinfo);

  if ( win == None ) {
    std::cerr << "no window selected\n";
    exit(1);
  } else {

    std::string classname;
    std::string classclass;
    std::string name;

    XClassHint ch;

    if (XGetClassHint(fbinfo.display, win, &ch) == 0) {
      classname = "";
      classclass = "";
    } else {
      if (ch.res_class != 0) {
        classname= const_cast<char *>(ch.res_name);
        XFree(ch.res_name);
        ch.res_name = 0;
      } else
        classname= "";

      if (ch.res_class != 0) {
        classclass = const_cast<char *>(ch.res_class);
        XFree(ch.res_class);
        ch.res_class = 0;
      } else
        classclass = "";
    }

    char* tmp_name = 0;

    XFetchName(fbinfo.display, win, &tmp_name);
    if (!tmp_name) {
      name = "";
    } else {
      name = tmp_name;
      XFree(tmp_name);
    }

    if (classclass == "" || classname == "" || name == "") {
      result = "#!! you selected a nonvalid window !!\n";
    } else
      result += "#!! select one of the [app] - entries !!\n";

    if (classclass != "")
      result += "[app] (class=" + classclass + ")\n";
    if (classname != "")
      result += "[app] (name=" + classname + ")\n";
    if (name != "")
      result += "[app] (title=" + name + ")\n";
        
    ArgList::const_iterator it;
    std::string entry;
    for (it = list.begin(); it != list.end(); it++) {
      entry = toLower(*it);
      if (entry == "workspace") 
        result += "[workspace]  {EDIT: INT}\n";
      else if (entry == "layer")
        result += "[layer]  {EDIT: INT|LITERAL}\n";
      else if (entry == "sticky")
        result += "[sticky]  {EDIT: yes|no}\n";
      else if (entry == "shaded")
        result += "[shaded]  {EDIT: yes|no}\n";
      else if (entry == "deco")
        result += "[deco]  {EDIT: LITERAL}\n";
      else if (entry == "position" || entry == "pos")
        result += "[position]  {EDIT: INT INT}\n";
      else if (entry == "hide" || entry == "hidden")
        result += "[hidden]  {EDIT: yes|no}\n";
      else if (entry == "hidefocus" || entry == "focushidden")
        result += "[focushidden]  {EDIT: yes|no}\n";
      else if (entry == "hideicon" || entry == "iconhidden")
        result += "[iconhidden]  {EDIT: yes|no}\n";
    }
        
    result += "[end]\n";
  }

  return result;
}

int numberOfWorkspaces(const FbInfo& fbinfo) {
  Atom real_type;
  Atom ewmh_atom = XInternAtom(fbinfo.display,
                               "_NET_NUMBER_OF_DESKTOPS",
                               False);
  Atom gnome_atom = XInternAtom(fbinfo.display,
                               "_WIN_WORKSPACE_COUNT",
                               False);
  int format;
  unsigned long num_elements, not_used;
  unsigned char *data = 0;
  if (XGetWindowProperty(fbinfo.display,
                         fbinfo.root_win,
                         ewmh_atom, 0, 1,
                         False, 0,
                         &real_type, &format, &num_elements, &not_used,
                         (unsigned char**)&data) == Success) {
      if (data)
        return static_cast<int>(*data);

  } else if (!data && XGetWindowProperty(fbinfo.display,
                         fbinfo.root_win,
                         gnome_atom, 0, 1,
                         False, 0,
                         &real_type, &format, &num_elements, &not_used,
                         (unsigned char**)&data) == Success) {
      if (data)
        return static_cast<int>(*data);
  }                      

  return 0;
}

int numberOfCurrentWorkspace(const FbInfo& fbinfo) {
  Atom real_type;
  Atom ewmh_atom = XInternAtom(fbinfo.display,
                               "_NET_CURRENT_DESKTOP",
                               False);
  Atom gnome_atom = XInternAtom(fbinfo.display,
                                "_WIN_WORKSPACE",
                                False);
                                
  int format;
  unsigned long num_elements, not_used;
  unsigned char *data = 0;
  if (XGetWindowProperty(fbinfo.display,
                         fbinfo.root_win,
                         ewmh_atom, 0, 1,
                         False, 0,
                         &real_type, &format, &num_elements, &not_used,
                         (unsigned char**)&data) == Success) {
      if (data)
        return static_cast<int>(*data);
  } else if (!data && XGetWindowProperty(fbinfo.display,
                         fbinfo.root_win,
                         gnome_atom, 0, 1,
                         False, 0,
                         &real_type, &format, &num_elements, &not_used,
                         (unsigned char**)&data) == Success) {
      if (data)
        return static_cast<int>(*data);
  }

  
  return -1;
}

std::vector< std::string > namesOfWorkspaces(const FbInfo& fbinfo) {

  std::vector < std::string > names;

  XTextProperty text_prop;
  char** tmp_names = 0;
  int   tmp_nr = 0;
  Atom atom = XInternAtom(fbinfo.display,
                          "_NET_DESKTOP_NAMES",
                          False);

  XGetTextProperty(fbinfo.display, fbinfo.root_win,
                   &text_prop, atom);
  XTextPropertyToStringList(&text_prop, &tmp_names, &tmp_nr);

  for (int i = 0; i < tmp_nr; i++ ) {
    names.push_back(tmp_names[i]);
  }
  XFreeStringList(tmp_names);

  return names;
}

void changeWorkspace(const FbInfo& fbinfo, int workspace) {
  Atom atom = XInternAtom(fbinfo.display,
                          "_BLACKBOX_CHANGE_WORKSPACE",
                           False);
  sendMessageToRootWindow(fbinfo, atom, workspace);
}

int isApp(const std::string& name) {

  PathList paths;

  stringToK< PathList >(paths, getenv("PATH"), ":");
  struct stat fstatus;

  for( PathList::iterator it= paths.begin();
       it != paths.end();
       it++ ) {

    if (stat((*it + "/" + name).c_str(), &fstatus) == 0 &&
        S_ISREG(fstatus.st_mode) &&
        ( fstatus.st_mode & S_IXUSR ||
          fstatus.st_mode & S_IXGRP ||
          fstatus.st_mode & S_IXOTH) ) {
      return 1;
    }
  }

  return 0;
}

int getPid(const FbInfo& fbinfo) {
  Atom real_type;
  Atom atom = XInternAtom(fbinfo.display,
                          "_BLACKBOX_PID",
                          False);
  int format;
  unsigned long num_elements, not_used;
  unsigned long* data = 0;
  if (XGetWindowProperty(fbinfo.display,
                         fbinfo.root_win,
                         atom, 0, 1,
                         False, 0,
                         &real_type, &format, &num_elements, &not_used,
                         (unsigned char**)&data) == Success) {
      if (data != 0) {
        return static_cast<int>(*data);
      }
  }

  return 0;
}

/*------------------------------------------------------------------*\
\*------------------------------------------------------------------*/

int cmdHelp(const FbInfo& fbinfo, const ArgList& arglist) {
  displayUsage();
  return 0;
}

int cmdVersion(const FbInfo& fbinfo, const ArgList& arglist) {
  std::cerr << "fbcmd version " << VERSION << "\n";
  return 0;
}


int cmdWorkspace(const FbInfo& fbinfo, const ArgList& arglist) {

  if (arglist.empty()) {
    std::cerr << "fbcmd: Error: missing argument for [workspace].\nExiting...\n";
    return 1;
  }

  int workspace = atoi(arglist[0].c_str());
  int current = numberOfCurrentWorkspace(fbinfo);

  if ( workspace == current )
    return 0;

  int max_workspaces = numberOfWorkspaces(fbinfo);

  if ( workspace < 0 || workspace >= max_workspaces ) {
    std::cerr << "fbcmd: Error: wrong range for [workspace].\nExiting...\n" <<
                 "possible : " << 0 << " <= id < " << max_workspaces <<
                 ". exit.\n";
    return 1;
  }

  changeWorkspace(fbinfo, workspace);

  return 0;
}

int cmdNextWorkspace(const FbInfo& fbinfo, const ArgList& arglist) {

  int delta = 1;

  if (!arglist.empty() && atoi(arglist[0].c_str()) > 0 ) {
    delta = atoi(arglist[0].c_str());
  }

  int current = numberOfCurrentWorkspace(fbinfo);
  int max_workspaces = numberOfWorkspaces(fbinfo);

  changeWorkspace(fbinfo, (current + delta) % max_workspaces);

  return 0;
}

int cmdPrevWorkspace(const FbInfo& fbinfo, const ArgList& arglist) {

  int delta = 1;

  if (!arglist.empty() && atoi(arglist[0].c_str()) > 0 ) {
    delta = atoi(arglist[0].c_str());
  }

  int max_workspaces = numberOfWorkspaces(fbinfo);
  int current = numberOfCurrentWorkspace(fbinfo);

  changeWorkspace(fbinfo, (((current - delta) % max_workspaces) + max_workspaces) % max_workspaces);

  return 0;
}

int cmdPid(const FbInfo& fbinfo, const ArgList& arglist) {

  std::cerr << getPid(fbinfo) << "\n";
  return 1;
}

int cmdPrint(const FbInfo& fbinfo, const ArgList& arglist) {

  if ( arglist.empty() || arglist[0] == "help" ) {
    std::cerr << "fbcmd: Error: missing arguments for [print], "
      << "possible arguments:\n"
      << "  nr_workspaces | nr_ws \n"
      << "  workspace | ws\n"
      << "  pid\n"
      << "  workspace_name | ws_name\n"
      << "  workspace_names | ws_names\n"
      << "  apps_entry\n";

    return 1;
  }

  if ( arglist[0] == "pid" ) {
    return cmdPid(fbinfo, arglist);
  } else if ( arglist[0] == "workspace" ||
              arglist[0] == "ws" ) {
    std::cout << numberOfCurrentWorkspace(fbinfo) << "\n";
  } else if ( arglist[0] == "nr_workspaces" ||
              arglist[0] == "nr_ws" ) {
    std::cout << numberOfWorkspaces(fbinfo) << "\n";
  } else if ( arglist[0] == "workspace_name" ||
              arglist[0] == "ws_name" ) {
    std::cout << namesOfWorkspaces(fbinfo)[numberOfCurrentWorkspace(fbinfo)] << "\n";
  } else if ( arglist[0] == "workspace_names" ||
              arglist[0] == "ws_names" ) {
    std::vector<std::string> names = namesOfWorkspaces(fbinfo);
    for(unsigned int i = 0; i < names.size(); i++ )
      std::cout << names[i] << " ";
    std::cout << "\n";
  } else if ( arglist[0] == "apps_entry" ) {
    std::cout << printAppsEntry(fbinfo, arglist);
  }

  return 0;
}

int cmdLogout(const FbInfo& fbinfo, const ArgList& arglist) {

  if ( arglist.empty() || arglist[0] == "help" ) {
    std::cerr << "fbcmd: Error: missing arguments for [logout], "
      << "possible arguments:\n"
      << "  xmessage [message]\n"
      << "  gxmessage [message]\n"
      << "  kdialog [message]\n"
      << "  gdialog [message]\n"
      << "  zenity [message]\n"
      << "  \"/path/to/binary --with=options\" [message]\n";

    return 1;
  }

  int pid = getPid(fbinfo);

  if ( pid <= 0 ) {
    std::cerr << "fbcmd: Error: couldnt determine the pid of fluxbox.\nExiting...\n";
    return 1;
  }

  std::string logout = (arglist.size() < 1 ?
                        "xmessage" :
                        arglist[0]);

  std::string options = "";

  for( unsigned int i = 0; logouts[i].appname; i++ ) {
    if ( logouts[i].appname == logout ) {
      if ( isApp(logout) ) {
        options = logouts[i].options;
      } else {
        std::cerr << "fbcmd: Error: " << logout << " doesn't exist in the PATH. Make sure the program is installed.\nExiting...\n";
        return 1;
      }
      break;
    }
  }

  std::string message = (arglist.size() < 2 ?
                         "do you really want to logout?\n" :
                         arglist[1]);

  std::ostringstream cmdline;

  cmdline << logout << options
    << "\"" << message << "\""
    << "; if [ $? -eq 0 ]; then kill -TERM " << pid << ";fi";

  system(cmdline.str().c_str());

  return 0;
}


/*------------------------------------------------------------------*\
\*------------------------------------------------------------------*/
int main(int argc, char* argv[]) {

  unsigned int i;

  if ( argc < 2 ) {
    displayUsage();
    exit(0);
  }

  std::string cmd = toLower(argv[1]);

  if ( cmd == "help" ||
       cmd == "--help" ||
       cmd == "-h" ||
       cmd == "-?" ) {
    displayUsage();
    exit(0);
  }

  ArgList arglist;
  for (i = 2; i < static_cast<unsigned int>(argc); i++ ) {
    arglist.push_back(argv[i]);
  }

  if ( !(getenv("DISPLAY")) ) {
    std::cerr << "fbcmd: Error: no DISPLAY set.\nExiting...\n";
    exit(1);
  }

  int ret = 1;
  FbInfo fbinfo;
  fbinfo.display = XOpenDisplay(getenv("DISPLAY"));

  if ( fbinfo.display ) {

    fbinfo.screen_number = DefaultScreen(fbinfo.display);
    fbinfo.root_win = RootWindow(fbinfo.display, fbinfo.screen_number);

    for(i = 0; fbcommands[i].cmdname; i++ ) {

      if ( fbcommands[i].cmdname == cmd ) {
        ret = fbcommands[i].command(fbinfo, arglist);
        break;
      }
    }

    XCloseDisplay(fbinfo.display);
  } else {
    std::cerr << "couldnt open display\n";
    exit(1);
  }

  return ret;
}

/* --------------------------------------------------------------- *\
\* --------------------------------------------------------------- */


